#!perl -w
#
# ModDepTab -- encapsulate modprobe(8)
#   Copyright (C) 2005  Erik van Konijnenburg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Most important routine is resolve(), which returns a list of
# modules with dependencies resolved.
#
# For every module, we have a list of prerequisite modules;
# these prerequisites are in no particular order: the prerequisites
# may have dependencies among themselves that are not satisfied
# by the order in the list.
#
# This emulates modprobe(8) by interpreting /lib/modules/2.6.x/modules.dep.
# Input format:
#  - characters are bytes, no messing with utf-8.
#  - backslash at end of line merges lines
#  - other \x get replaced by x
#  - ^\s*# lines are ignored
#  - lines without : are ignored.
# This means "aap: noot.#7" is a valid dependency.
# The backslash interpretation is mostly for modprobe.conf;
# depmod does not generate it.
#
# Modprobe determines module name by dropping everything after dot:
# "/lib/noot.#7" is module "noot".  We'll adopt the same policy.
#
# Note that depmod only grabs modules ending in .ko or .ko.gz;
# we'll issue a warning if a module has a different suffix.
#
# Note that modprobe does not discriminate against modules outside
# /lib/modules: if it's listed in modules.dep, it's a valid module.
#
# Note that redhat has a convention that modules in .../update take
# precedence over other modules with the same name.  Depmod implements
# this by not putting modules that are overridden in modules.dep.
# Thus modprobe and yaird need no special action to support that convention.
#
# The logic modprobe uses to determine which modules to load:
# - replace hyphens with underscore
# - read config specified on command line,
#   OR modprobe.conf OR modprobe.d, only first one found.
#   Remember all "install" and "option" directives,
#   rewrite module name if an alias matches
# - if no alias found and the name is of the form 'symbol:xxx':
# 	look in modules.symbols to resolve to a module name
# - if alias found:
#	make a list of modules to load, based on modules.dep
# - else:
#	make a list of modules to load, based on modules.dep
#	if that turned up no modules AND there was no install cmd:
#		- look in modules.aliases to resolve to modulename
#		- make a list of modules to load, based on modules.dep
# - if the list to load is empty AND there was no install command:
# 	- complain.
# # in insmod():
# - recurse over the module list, most basic stuff first, doing:
# 	- if there is a command for this module name:
# 		execute it
# 	- else:
#		load the module
#
# We probably should replace this perl module with parsing the
# output of modprobe --verbose --dry-run --show-depends --set-version,
# but that requires changes in the calling modules, and there's discussion
# of adding blacklist support to modprobe, so for now, let's not.
#

use strict;
use warnings;
use Base;
use Conf;
use ModDep;
use KConfig;

package ModDepTab;

my $modDepTab = undef;
my $modDepList = undef;

sub init () {
	if (defined ($modDepTab)) {
		return;
	}
	$modDepTab = {};
	my $name = Conf::get('modDep');
	if (! open (IN, "<", "$name")) {
		Base::fatal ("can't open $name");
	}
	my $lineNo = 0;
	while (defined (my $line = <IN>)) {
		$lineNo++;
		chomp $line;

		# it would be nice to ignore leading and trailing
		# space, and to allow # comment at end of line.
		# however, that's not how modprobe does it,
		# so we shouldn't either.
		next if $line =~ /^\s*#/;

		if ($line =~ /^(.*):\s*(.*)$/) {
			my $file = $1;
			my $val = $2;
			my $key = $file;

			if ($key !~ /\.ko$/  && $key !~ /\.ko\.gz/) {
				Base::warning ("$name:$lineNo: odd module name: $key");
			}
			$key =~ s!.*/!!;
			$key =~ s!\..*$!!;
			$key =~ s!_!-!g;

			my @fields = split (/\s+/, $val);
			for my $i (0 .. $#fields) {
				# strip pathname and extension from
				# required modules, normalise underscore.
				# dont complain about odd suffix.
				$fields[$i] =~ s!.*/!!;
				$fields[$i] =~ s!\..*$!!;
				$fields[$i] =~ s!_!-!g;
			}
			
			my $conflict = $modDepTab->{$key};
			if (defined ($conflict)) {
				my $origin = $conflict->origin();
				Base::warning ("$name:$lineNo: $key overrules earlier definition at $origin");
			}

			$modDepTab->{$key} = ModDep->new (
				name => $key,
				file => $file,
				origin => "$name:$lineNo",
				deps => [ @fields ],
				);
		}
		else {
			Base::warning ("$name:$lineNo: malformed line ignored");
		}
		
	}
	if (! close (IN)) {
		Base::fatal ("could not read $name");
	}
	$modDepList = [ sort keys %{$modDepTab} ];
}

sub all	() {
	init;
	return $modDepList;
}

sub findByModnam ($) {
	my ($module) = @_;
	init;
	$module =~ s!_!-!g;
	my $result = $modDepTab->{$module};
	if (! defined ($result)) {
		Base::fatal ("can't find dependencies for $module");
	}
	return $result;
}


# return a copy of the list with prerequisite modules inserted
sub resolve ($) {
	my ($modList) = @_;
	Base::debug ("resolving: " . join (',', @{$modList}) . '.');
	my $result = [];
	resolve2 ($modList, $result);
	Base::debug ("resolved to: " . join (',', @{$result}) . '.');
	return $result;
}

#
# resolve2 -- given a list of modules, that may have dependencies
# (even among themselves) and a list of modules that is in proper order
# to be loaded without unresolved dependencies,
# add everything from the first list (plus their dependencies)
# to the second list, without introducing duplicates.
#
sub resolve2 ($$);
sub resolve2 ($$) {
	my ($modList, $done) = @_;
OUTER:	for my $module (@{$modList}) {
		#
		# Stuff that is compiled into the kernel cannot
		# add or resolve dependencies, and obviously don't
		# need to be loaded.  Drop them from the list.
		#
		if (KConfig::isBuiltIn ($module)) {
			next OUTER;
		}
		#
		# if the module is already there, don't add again
		#
		for my $d (@{$done}) {
			if ($d eq $module) {
				next OUTER;
			}
		}
		#
		# recursively add prerequisites
		#
		my $deps = ModDepTab::findByModnam($module);
		if (! defined ($deps)) {
			Base::fatal ("can't find dependencies for $module");
		}
		resolve2 ($deps->deps, $done);
		#
		# and module itself
		#
		push @{$done}, $module;
	}
}

1;
